from abc import ABCMeta

class Variable(metaclass=ABCMeta):
    def __init__(self, label: str):
        self._label = label
        self.var_label = label

    @property
    def label(self):
        return self._label

    def set_relabel(self, label: str):
        self._label = label



from typing import Dict, Type

from numpy.lib.function_base import place

from jijmodeling.variables.to_pyqubo import ToPyQUBO
from jijmodeling.express.express import Express
from jijmodeling.variables.variable import Variable
import pyqubo
import numpy as np


class Placeholder(Express, Variable, ToPyQUBO):
    """This class represents placeholder for lazy evaluation.

    Args:
        Express ([type]): [description]
        Variable ([type]): [description]
        ToPyQUBO ([type]): [description]
    """
    def __init__(self, label: str):
        self._label = label
        super().__init__([])

    def __repr__(self):
        return self._label

    @property
    def label(self):
        return self._label

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        label = seriablizable['attributes']['_label']
        return cls(label=label)

    def to_pyqubo(self, index: Dict[str, int]={}, placeholder: dict={}, fixed_variables: dict={}) -> pyqubo.Express:

        if self.label in placeholder:
            # placeholder value is not Express
            if isinstance(placeholder[self.label], Express):
                raise TypeError("placeholder's value is not Express class.")
            return placeholder[self.label]


        label_list = self.label.split('[')
        label = label_list[0]
        indices = [int(l[:-1]) for l in label_list[1:]]
        if label in placeholder:
            if isinstance(placeholder[label], np.ndarray):
                return placeholder[label][tuple(indices)]
            else:
                return placeholder[label]
        else:
            return pyqubo.Placeholder(label=self.label)



from typing import Union, Tuple

from jijmodeling.variables import Binary, DisNum, Placeholder
from jijmodeling.array.array import Array, ArraySizePlaceholder

ShapeType = Union[int, ArraySizePlaceholder, Tuple[Union[int, ArraySizePlaceholder], ...]]

def BinaryArray(label: str, shape: ShapeType):
    return Array(Binary(label), shape)


def DisNumArray(label: str, shape: ShapeType, lower: float=0.0, upper: float=1.0, bits: int=3):
    return Array(DisNum(label, lower, upper, bits), shape)


def PlaceholderArray(label: str, shape: ShapeType=None, dim: int=None):
    if shape is None and isinstance(dim, int):
        # _shape = tuple([ArraySizePlaceholder(label=label + '_shape_%d' % i) for i in range(dim)])
        # array = Array(Placeholder(label), _shape)
        # for i, s in enumerate(_shape):
        #     s.set_array(array, i)
        _shape = tuple(None for _ in range(dim))
        return Array(Placeholder(label), _shape)
        return array
    elif shape is not None:
        shape = (shape, ) if isinstance(shape, int) else shape
        return Array(Placeholder(label), shape)
    else:
        raise ValueError("Input shape or dim.")
    



from typing import List, Any, Union, Dict
from abc import ABCMeta

import pyqubo
from jijmodeling.variables.to_pyqubo import ToPyQUBO, to_pyqubo


class Express(metaclass=ABCMeta):
    """All component inheritance this class.
    This class provide computation rules.
    """
    def __init__(self, children: List[Any]):
        self.children = children
        index_labels = []
        for child in self.children:
            if isinstance(child, Express):
                index_labels += child.index_labels
        self.index_labels = list(set(index_labels))

    def __sub__(self, other):
        return self.__add__(-1*other)

    def __rsub__(self, other):
        return Add(other, -1*self)

    def __lsub(self, other):
        return self.__add__(-1*other)
    
    def __add__(self, other):
        return Add(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __ladd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        return Mul(self, other)
    
    def __lmul__(self, other):
        return self.__mul__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, other):
        return Power(self, other)

    def __repr__(self):
        str_repr = ""
        for child in self.children:
            str_repr +=  child.__repr__() + " "
        return str_repr[:-1]

    def to_serializable(self):
        def express_to_ser(s):
            if 'to_serializable' in dir(s):
                return s.to_serializable()
            elif isinstance(s, (list, tuple)):
                return {
                    'iteratable': 'list' if isinstance(s, list) else 'tuple',
                    'value': [express_to_ser(t) for t in s],
                }
            elif isinstance(s, dict):
                return {k: express_to_ser(v) for k,v in s.items()}
            else:
                return s
        serializable = {
            'class': self.__class__.__module__ + '.' + self.__class__.__name__,
            'attributes': {k: express_to_ser(v) for k, v in vars(self).items()}
        }
        return serializable

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        _cls_serializable_validation(seriablizable, cls)
        children = from_serializable(seriablizable['attributes']['children'])
        return cls(children)


TermType = Union[Express, int, float]
        

class Add(Express, ToPyQUBO):
    def __init__(self, *terms: TermType):
        super().__init__(list(terms))

    def __add__(self, other):
        self.children.append(other)
        return self

    def __repr__(self):
        str_repr = ""
        for t in self.children:
            str_repr += t.__repr__() + ' + '
        return str_repr[:-3]


    def to_pyqubo(self, index: Dict[str, int]={}, placeholder: dict={}, fixed_variables: dict={}) -> pyqubo.Express:

        return sum(to_pyqubo(child, index=index, placeholder=placeholder, fixed_variables=fixed_variables) for child in self.children)

    def value(self, solution, placeholder={}, index={}):
        def to_value(term):
            if isinstance(term, Express):
                return term.value(solution, placeholder, index)
            else:
                return term
        return sum(to_value(child) for child in self.children)

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        children = from_serializable(seriablizable['attributes']['children'])
        return cls(*children)

class Mul(Express, ToPyQUBO):
    def __init__(self, *terms: TermType):
        super().__init__(list(terms))

    def __mul__(self, other):
        self.children.append(other)
        return self

    def __repr__(self):
        str_repr = ""
        for t in self.children:
            if isinstance(t, Add) and len(t.children) > 1:
                str_repr += '(%s)' % t.__repr__()
            if isinstance(t, (int, float)):
                str_repr += '({})'.format(t.__repr__())
            else:
                str_repr += t.__repr__()
        return str_repr

    def to_pyqubo(self, **kwargs) -> pyqubo.Express:
        term = 1
        for child in self.children:
            term *= child.to_pyqubo(**kwargs) if isinstance(child, ToPyQUBO) else child
        return term

    def value(self, solution, placeholder, index):
        term = 1
        def to_value(term):
            if isinstance(term, Express):
                return term.value(solution, placeholder, index)
            else:
                return term

        for child in self.children:
            term *= to_value(child)
        return term 

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        children = from_serializable(seriablizable['attributes']['children'])
        return cls(*children)

class Power(Express, ToPyQUBO):
    def __init__(self, base: TermType, exponent: int):
        super().__init__([base, exponent])
        self.base = base
        self.exponent = exponent

    def to_pyqubo(self, **kwargs) -> pyqubo.Express:

        term = to_pyqubo(self.base, **kwargs)

        return term ** self.exponent


    def __repr__(self) -> str:
        return '({})^{}'.format(self.base, self.exponent)


    def value(self, solution, placeholder={}, index={}):
        value = self.base.value(solution, placeholder, index)
        return value ** self.exponent

    @classmethod
    def from_serializable(cls, seriablizable: dict):

        base = from_serializable(seriablizable['attributes']['base'])
        exponent = from_serializable(seriablizable['attributes']['exponent'])
        return cls(base, exponent)



def from_serializable(serializable):

    if isinstance(serializable, dict) and 'class' in serializable:
        import jijmodeling
        modulePath = serializable['class'].split('.')[2:]
        module = jijmodeling
        for m in modulePath:
            module = getattr(module, m)
        _cls_serializable_validation(serializable, module)
        return module.from_serializable(serializable)

    elif isinstance(serializable, dict) and 'iteratable' in serializable:
        if serializable['iteratable'] == 'list':
            return [from_serializable(s) for s in serializable['value']]
        elif serializable['iteratable'] == 'tuple':
            return tuple(from_serializable(s) for s in serializable['value'])
    elif isinstance(serializable, list):
        return [from_serializable(s) for s in serializable]
    elif isinstance(serializable, dict):
        return {k: from_serializable(v) for k, v in serializable.items()}
    else:
        return serializable

def _cls_serializable_validation(serializable, cls):
    if 'class' not in serializable:
        return
    class_name = serializable['class'].split('.')[-1]
    if cls.__name__ != class_name:
        raise ValueError('Class "{}" is not class "{}".'.format(cls.__name__, class_name))


import pyqubo

from jijmodeling.variables.to_pyqubo import ToPyQUBO
from jijmodeling.express.express import Express, from_serializable

class Constraint(Express, ToPyQUBO):
    def __init__(self, label: str, term, condition='==', constant=0, with_mul=True):
        super().__init__([term])
        self.label = label
        self.condition = condition
        self.constant = constant
        self.with_mul = with_mul

    def to_pyqubo(self, **kwargs) -> pyqubo.Express:
        term = self.children[0].to_pyqubo(**kwargs)
        if self.with_mul:
            return pyqubo.Placeholder(self.label) * pyqubo.Constraint(term, self.label)
        return pyqubo.Constraint(term, self.label)

    def value(self, solution, placeholder):
        return self.children[0].value(solution, placeholder)

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        term = from_serializable(seriablizable['attributes']['children'])[0]
        label = from_serializable(seriablizable['attributes']['label'])
        condition = from_serializable(seriablizable['attributes']['condition'])
        constant = from_serializable(seriablizable['attributes']['constant'])
        with_mul = from_serializable(seriablizable['attributes']['with_mul'])
        return cls(label, term, condition, constant, with_mul)


from typing import List, Tuple, Union, Type, Dict
import numbers

from numpy.lib.function_base import place
from pyqubo import array

from jijmodeling.variables.variable import Variable
from jijmodeling.variables.to_pyqubo import ToPyQUBO, to_pyqubo
from jijmodeling.variables.binary import Binary
from jijmodeling.variables.placeholder import Placeholder
from jijmodeling.sum import Sum
from jijmodeling.express.express import Express, from_serializable
import pyqubo
import numpy as np


class Tensor(Express, ToPyQUBO):
    """Tensor class for each variables

    Args:
        label
    """

    def __init__(self, label: str, variable, shape: tuple, indices: List[str]):
        """Tensor class for each variables

        Args:
            label (str): label for variable
            variable ([type]): variable object
            shape (tuple): shape of variable tensor
            indices (List[str]): index labels
        """
        super().__init__([])
        self.label = label
        self.variable = variable
        self.index_labels = indices
        self.shape = (shape, ) if isinstance(shape, int) else shape

    def __repr__(self):
        index_str = ''
        for i in self.index_labels:
            index_str += '[%s]' % i
        return self.label + index_str

    def to_pyqubo(self, index: Dict[str, int]={}, placeholder: dict={}, fixed_variables: dict={}) -> Union[pyqubo.Express, numbers.Number]:
        index_label = self.label

        for label in self.index_labels:
            if not isinstance(label, str):
                _label = to_pyqubo(label, index=index, placeholder=placeholder, fixed_variables=fixed_variables)
                index_label += '[%d]' % _label
            else:
                index_label += '[%d]' % index[label]
        
        if index_label in fixed_variables:
            return fixed_variables[index_label]

        self.variable.set_relabel(index_label)
        return self.variable.to_pyqubo(index=index, placeholder=placeholder, fixed_variables=fixed_variables)

    def value(self, solution, placeholder, index):
        if self.label in solution:
            sol = solution[self.label]
        elif self.label in placeholder:
            sol = placeholder[self.label]

        def to_index(obj):
            if isinstance(obj, str):
                return index[obj]
            else:
                return to_pyqubo(obj, placeholder=placeholder)
        index_label = [to_index(label) for label in self.index_labels]
        return sol[tuple(index_label)]

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        label = from_serializable(seriablizable['attributes']['label'])
        variable = from_serializable(seriablizable['attributes']['variable'])
        shape = from_serializable(seriablizable['attributes']['shape'])
        indices = from_serializable(seriablizable['attributes']['index_labels'])
        return cls(label, variable, tuple(shape), indices)






class ArraySizePlaceholder(Placeholder):

    def __init__(self, label:str) -> None:
        super().__init__(label)
        self._array = None

    @property
    def array(self):
        return self._array

    def set_array(self, array, shape_index: int):
        self._array = array
        self._shape_index = shape_index
    
    def to_pyqubo(self, index: Dict[str, int]={}, placeholder: dict={}, fixed_variables: dict={}) -> pyqubo.Express:
        array_data = placeholder[self._array.label]
        if isinstance(array_data, Express):
            raise TypeError('Placeholder "{}" should not be Express class.'.format(self._array.label))

        if isinstance(array_data.shape[self._shape_index], Express):
            raise TypeError('Placeholder "{}" should not be Express class.'.format(self._array.label))
        return array_data.shape[self._shape_index]

    def __eq__(self, other):
        if not isinstance(other, ArraySizePlaceholder):
            return False
        return self.label == self.label

    def __hash__(self):
        return hash(self.label)

    def to_serializable(self):

        return {
            'class': self.__class__.__module__ + '.' + self.__class__.__name__,
            'attributes': {
                'children': [],
                'label': self.label,
                '_array': self._array.to_serializable(),
                '_shape_index': self._shape_index
            },
        }

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        placeholder = cls(seriablizable['attributes']['label'])
        placeholder.set_array(
            array=Array.from_serializable(seriablizable['attributes']['_array']),
            shape_index=seriablizable['attributes']['_shape_index']
        )
        return placeholder




class Array(Variable):
    """Tensor object generator
    This class generate Tensor class by index access

    """
    def __init__(self, variable, shape: Union[int, tuple, ArraySizePlaceholder, None]):
        self.variable = variable
        self._label = variable.label
        shape_list = list(shape) if isinstance(shape, tuple) else [shape] 
        for i, s in enumerate(shape_list):
            if isinstance(s, ArraySizePlaceholder):
                if s.array is None:
                    s.set_array(self, i)
                shape_list[i] = s
        self._shape = shape_list

        self.dim = len(self._shape)

    def to_serializable(self):
        def express_to_ser(s):
            if 'to_serializable' in dir(s):
                return s.to_serializable()
            elif isinstance(s, list):
                return [express_to_ser(t) for t in s]
            else:
                return s
        
        # shape が　ArraySizePlaceholder の場合,
        # このクラスへの参照を持つのでそれは避けるようにする
        variables = vars(self)
        variables['shape'] = self._shape


        serializable = {
            'class': self.__class__.__module__ + '.' + self.__class__.__name__,
            'attributes': {k: express_to_ser(v) for k, v in variables.items()}
        }

        return serializable

    @classmethod
    def from_serializable(cls, serializable):
        variable = from_serializable(serializable['attributes']['variable'])
        shape = from_serializable(serializable['attributes']['shape'])
        return cls(variable, shape)
    
    @property
    def shape(self):
        shape = []
        for i, s in enumerate(self._shape):
            if s is None:
                asph = ArraySizePlaceholder(label=self.label + '_shape_%d' % i)
                asph.set_array(self, i)
                shape.append(asph)
            else:
                shape.append(s)
        return tuple(shape)


    def __getitem__(self, key: Union[str, slice, Tuple[Union[str, slice], ...]])->Type[Express]:
        """Gerate Tensor class

        Args:
            key (Union[str, slice, Tuple[Union[str, slice], ...]]): [description]

        Raises:
            ValueError: [description]

        Returns:
            Type[Express]: [description]
        """
        if not isinstance(key, tuple):
            key = (key, )

        if len(key) != self.dim:
            raise ValueError("{}'s dimension is {}.".format(self.label, self.dim))

        indices: List[str] = []
        summation_index = []
        for i, k in enumerate(list(key)):
            # for syntaxsugar x[:]
            if isinstance(k, slice):
                indices.append(':_{}'.format(i))
                summation_index.append((i, indices[i]))
            elif isinstance(k, (int, str, Express)):
                indices.append(k)

        term = Tensor(self.label, self.variable, self.shape, indices)

        # for syntaxsugar x[:]
        for i, ind in summation_index:
            term = Sum({ind: self.shape[i]}, term)

        return term

    def decode_dimod_response(self, response, placeholder)->dict:
        # decode shape
        _shape = tuple([to_pyqubo(s, placeholder=placeholder) for s in self.shape])
        return self.variable.decode_dimod_response(response, shape=_shape)


from typing import List, Union, Tuple, Type

import numpy as np
import pyqubo

from jijmodeling.express.express import Express
from jijmodeling.express.express import from_serializable
from jijmodeling.variables.to_pyqubo import ToPyQUBO, to_pyqubo

class Sum(Express, ToPyQUBO):
    def __init__(self, indices: dict, term: Union['Sum', Type[Express]]):
        super().__init__(children=[term])
        self.indices = indices

        index_keys = list(indices.keys())
        self.index_labels = [ind for ind in self.index_labels if ind not in index_keys]

    def to_pyqubo(self, index: dict={}, placeholder: dict={}, fixed_variables: dict ={})->Union[int, float, pyqubo.Express]:
        term = self.children[0]
        sum_index = _reshape_to_index_set(self.indices, index, placeholder)
        return sum(term.to_pyqubo(index=ind, placeholder=placeholder, fixed_variables=fixed_variables) for ind in sum_index)

    def __repr__(self):
        repr_str = 'Σ_{'
        for i in self.indices.keys():
            repr_str += i + ', '
        term = self.children[0]
        repr_str = repr_str[:-2] + '}}({})'.format(term.__repr__()) 
        return repr_str

    def value(self, solution, placeholder={}, index={}):
        sum_index = _reshape_to_index_set(self.indices, index, placeholder)
        term = 0
        for child in self.children:
            for ind in sum_index:
                if isinstance(child, Express):
                    child_val = child.value(solution, placeholder, ind)
                    term += child_val if child_val is not None else 0
                else:
                    term += child
        return term

    @classmethod
    def from_serializable(cls, seriablizable: dict):
        indices = from_serializable(seriablizable['attributes']['indices'])
        term = from_serializable(seriablizable['attributes']['children'])[0]
        return cls(indices, term)
        




def _reshape_to_index_set(indices: dict, assigned_ind: dict, placeholder)->List[dict]:
    from jijmodeling.array import ArraySizePlaceholder
    index_lists = {}
    for label, V in indices.items():
        # TODO: corresponds to list of ArraySizePlaceholder
        if isinstance(V, (list, np.ndarray)):
            index, index_set = _satisfied_index_set(label, V, assigned_ind)
        elif isinstance(V, (int, ArraySizePlaceholder, Express)):
            V = to_pyqubo(V, placeholder=placeholder)
            index, index_set = _satisfied_index_set(label, list(range(V)), assigned_ind)
        elif isinstance(V, tuple):
            v0 = to_pyqubo(V[0], placeholder=placeholder)
            v1 = to_pyqubo(V[1], placeholder=placeholder)
            index, index_set = _satisfied_index_set(label, list(range(v0, v1)), assigned_ind)
        else:
            raise TypeError('index type is not {}'.format(type(V)))
        index_lists[index] = index_set

    indices_dict = []
    keys = list(index_lists.keys())
    num_indices = len(index_lists[keys[0]])
    for i in range(num_indices):
        ind_dict = {label: index_lists[label][i] for label in keys}
        ind_dict.update(assigned_ind)
        indices_dict.append(ind_dict)

    return indices_dict


def _satisfied_index_set(index_str: str, index_set:list, assigned: dict)->Tuple[str, list]:
    """index のkey文字列が i < j のような条件演算を含むとき, 条件を満たすsetを書き出してくる

    Args:
        index_str (str): [description]
        index_set (list): [description]
        assigned (dict): [description]

    Returns:
        Tuple[str, list]: [description]
    """
    ind_chars = index_str.split(' ')
    if len(ind_chars) == 1:
        return index_str, index_set

    index, operator, right_ind = ind_chars
    return index, [j for j in index_set if eval('{} {} {}'.format(j, operator, assigned[right_ind]))]
    







from abc import ABCMeta, abstractmethod
from typing import Dict, Union
import numbers
import pyqubo


class ToPyQUBO(metaclass=ABCMeta):

    def __init__(self, label: str) -> None:
        self.label = label
        self.index_labels = []

    @abstractmethod
    def to_pyqubo(self, index: Dict[str, int]={}, placeholder: dict={}, fixed_variables: dict={})->pyqubo.Express:
        raise NotImplementedError()
        


def to_pyqubo(object, **kwargs)->Union[pyqubo.Express, numbers.Number]:
    if isinstance(object, ToPyQUBO):
        return object.to_pyqubo(**kwargs)
    elif isinstance(object, numbers.Number):
        return object
    else:
        raise TypeError('{} cannnot convert to PyQUBO object.'.format(object))


from typing import Dict
from jijmodeling.variables.to_pyqubo import ToPyQUBO
from jijmodeling.express.express import Express
from jijmodeling.variables.variable import Variable
import pyqubo
import numpy as np


class Binary(Express, Variable, ToPyQUBO):

    def __init__(self, label: str) -> None:
        super().__init__([])
        self._label = label
        self.var_label = label

    def __repr__(self) -> str:
        return self.label

    def to_pyqubo(self, index: Dict[str, int]={}, placeholder: dict={}, fixed_variables: dict={}):
        """Convert to pyqubo.Binary class

        Returns:
            pyqubo.Binary: pyqubo object
        """

        if self.label in fixed_variables:
            return fixed_variables[self.label]

        return pyqubo.Binary(self.label)

    def decode_dimod_response(self, response, shape=None) -> dict:
        var_value_dict = {}
        dim = len(shape)

        if shape is None:
            for sample in response.samples():
                var_value_dict[self.var_label] = sample[self.label]
        else:
            for sample in response.samples():
                value_tensor = np.full(shape, np.nan, dtype=np.int)
                for key, value in sample.items():
                    label_list = key.split('[')
                    res_label = label_list[0]
                    if res_label == self.var_label:
                        indices = [int(k[:-1]) for k in label_list[-1*dim:]]
                        value_tensor[tuple(indices)] = value
                var_value_dict[self.var_label] = value_tensor

        return var_value_dict

            


from jijmodeling.array.array import Array
from jijmodeling.express import Constraint
from typing import TypeVar, Generic

Var = TypeVar('T')

class Problem(Generic[Var]):
    def __init__(self, problem_name: str) -> None:
        self.variables = []
        self.name = problem_name
        self.model = 0
        self.cost = 0
        self.constraints = {}

    def var(self, variable: Var)->Var:
        # if not isinstance(variable, Variable):
        #     raise TypeError('variable should be jijmodeling.Variable')
        self.variables.append(variable)
        return variable

    
    def __repr__(self) -> str:
        return '{}: {}'.format(self.name, self.model)


    def __add__(self, other):
        if self.model == 0:
            self.model = other
        else:
            self.model += other

        if isinstance(other, Constraint):
            self.constraints[other.label] = other
        else:
            self.cost = other if self.cost == 0 else self.cost + other

        return self


    def to_pyqubo(self, index: dict={}, placeholder: dict={}, fixed_variables: dict={}):
        self._placeholder = placeholder
        self._fixed_variables = fixed_variables
        return self.model.to_pyqubo(index=index, placeholder=placeholder, fixed_variables=fixed_variables)

 
    def decode_dimod_resopnse(self, response):
        decoded = []
        for i in range(len(response)):
            res = response.slice(i, i+1)
            vars_value = {}
            for var in self.variables:
                vars_value.update(var.decode_dimod_response(res, self._placeholder))
            
            # replace fixed variables
            for label, fixed_val in self._fixed_variables.items():
                label_list = label.split('[')
                val_label = label_list[0]
                index = [int(i[:-1]) for i in label_list[1:]]
                vars_value[val_label][tuple(index)] = fixed_val
            decoded.append(vars_value)

        return decoded

    def calc_penalty(self, solution):
        penalties = {}
        for k, const in self.constraints.items():
            penalties[k] = const.value(solution, self._placeholder)
            if const.condition == '==' and penalties[k] == const.constant:
                penalties[k] = 0.0
            elif const.condition == '<=' and penalties[k] <= const.constant:
                penalties[k] = 0.0
            else:
                penalties[k] = const.value(solution, self._placeholder) - const.constant
        return penalties

    def calc_cost(self, solution):
        return self.cost.value(solution, self._placeholder)


    def from_serializable(self, serializable: dict):
        import jijmodeling
        className = serializable['class'].split('.')[1:]
        cls = getattr(jijmodeling, className)
        return cls.from_serializable(serializable)


from jijmodeling.variables.variable import Variable
from jijmodeling.variables.binary import Binary
from jijmodeling.variables.dis_num import DisNum
from jijmodeling.variables.placeholder import Placeholder


from jijmodeling.variables import placeholder
from jijmodeling.variables.variable import Variable
from jijmodeling.variables.to_pyqubo import ToPyQUBO
from jijmodeling.express.express import Express
import pyqubo


class DisNum(Express, Variable, ToPyQUBO):
    def __init__(self, label: str, lower: float=0.0, upper: float=1.0, bits: int=3):
        super().__init__([])
        self._label = label
        self.lower = lower
        self.upper = upper
        self.bits = bits

    def __repr__(self) -> str:
        return self.label

    def to_pyqubo(self, index:dict={}, placeholder:dict={}, fixed_variables: dict = {}):
        if self.label in fixed_variables:
            return fixed_variables[self.label]
        var_label = self.label + '[{}]'
        coeff = (self.upper - self.lower)/(2**self.bits - 1)
        return coeff * sum(2**i * pyqubo.Binary(var_label.format(i)) for i in range(self.bits)) + self.lower